"""
Copyright (c) 2019 The Cereja Project
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import ast
import gc
import math
import time
from importlib import import_module
import importlib
import sys
import types
import random
from typing import Any, Union, List, Tuple, Sequence, Iterable
import logging
import itertools
from copy import copy
import inspect

# Needed init configs
from cereja.config.cj_types import ClassType, FunctionType, Number

__all__ = ['CjTest', 'camel_to_snake', 'combine_with_all', 'fill', 'get_attr_if_exists',
           'get_implements', 'get_instances_of', 'import_string',
           'install_if_not', 'invert_dict', 'logger_level', 'module_references', 'set_log_level', 'time_format',
           'string_to_literal', 'rescale_values', 'Source', 'sample', 'obj_repr', 'truncate', 'type_table_of',
           'list_methods', 'can_do', 'chunk', 'is_iterable', 'is_sequence', 'is_numeric_sequence', 'clipboard']

logger = logging.getLogger(__name__)


def chunk(data: Sequence, batch_size: int = None, fill_with: Any = None, is_random: bool = False,
          max_batches: int = None) -> List:
    """

    e.g:
    >>> import cereja as cj

    >>> data = list(range(15))
    [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14]

    >>> cj.chunk(data, batch_size=4)
    [[0, 1, 2, 3], [4, 5, 6, 7], [8, 9, 10, 11], [12, 13, 14]]

    >>> cj.chunk(data, batch_size=4, is_random=True, fill_with=0)
    [[7, 2, 11, 4], [10, 6, 1, 13], [12, 9, 5, 0], [8, 3, 14, 0]]

    >>> data = {"key1": 'value1', "key2": 'value2', "key3": 'value3', "key4": 'value4'}

    >>> cj.chunk(data, batch_size=2,is_random=True)
    [{'key3': 'value3', 'key2': 'value2'}, {'key1': 'value1', 'key4': 'value4'}]

    @param data: Iterable data
    @param batch_size: number of items per batch
    @param fill_with: Any, but isn't valid for dict
    @param is_random: shuffle data
    @param max_batches: limit number of batches
    @return: list of batches
    """

    assert is_iterable(data), f"Chunk isn't possible, because value {data} isn't iterable."
    if batch_size is None and max_batches is None:
        return [data]

    data = list(data) if isinstance(data, (set, tuple, str, bytes, bytearray)) else copy(data)
    _dict_temp_keys = [] if not isinstance(data, dict) else list(data)
    if not batch_size or batch_size > len(data) or batch_size < 1:
        if isinstance(max_batches, (int, float)) and max_batches > 0:
            batch_size = len(data) // max_batches or len(data)
        else:
            batch_size = len(data)

    if is_random:
        if isinstance(data, dict):
            random.shuffle(_dict_temp_keys)
        else:
            random.shuffle(data)

    if max_batches is None:
        max_batches = len(data) // batch_size if len(data) % batch_size == 0 else len(data) // batch_size + 1

    batches = []
    for i in range(0, len(data), batch_size):

        if isinstance(data, dict):
            result = {key: data[key] for key in _dict_temp_keys[i: i + batch_size]}
        else:
            result = data[i:i + batch_size]
            if fill_with is not None and len(result) < batch_size:
                result += [fill_with] * (batch_size - len(result))
        batches.append(result)
        max_batches -= 1
        if not max_batches:
            break
    return batches


def _get_tkinter():
    try:
        from tkinter import Tk
    except ImportError:
        raise ValueError("Sorry. Isn't possible.")
    return Tk()


def clipboard() -> str:
    return _get_tkinter().clipboard_get()


def truncate(text: Union[str, bytes], k=4):
    """
    Truncate text.
    eg.:
    >>> import cereja as cj
    >>> cj.utils.truncate("Cereja is fun.", k=3)
    'Cer...'

    @param text: string or bytes
    @param k: natural numbers, default is 4
    """
    assert isinstance(text, (str, bytes)), TypeError(f"{type(text)} isn't valid. Expected str or bytes")
    if k > len(text) or k < 0:
        return text
    trunc_chars = '...' if isinstance(text, str) else b'...'
    return text[:k] + trunc_chars


def obj_repr(obj_, attr_limit=10, val_limit=3, show_methods=False, show_private=False, deep=3):
    try:
        if isinstance(obj_, (str, bytes)):
            return truncate(obj_, k=attr_limit)
        if isinstance(obj_, (bool, float, int, complex)):
            return obj_
        rep_ = []
        if deep > 0:
            for attr_ in dir(obj_):
                if attr_.startswith('_') and not show_private:
                    continue
                obj = obj_.__getattribute__(attr_)

                if isinstance(obj, (str, bool, float, int, complex, bytes, bytearray)):
                    rep_.append(f'{attr_} = {obj_repr(obj)}')
                    continue
                if callable(obj) and not show_methods:
                    continue

                if is_iterable(obj):
                    temp_v = []
                    for k in obj:
                        if isinstance(obj, dict):
                            k = f'{k}:{type(obj[k])}'
                        elif is_iterable(k):
                            k = obj_repr(k, deep=deep)
                            deep -= 1
                        else:
                            k = str(k)
                        temp_v.append(k)
                        if len(temp_v) == val_limit:
                            break
                    temp_v = ', '.join(temp_v)  # fix me, if bytes ...
                    obj = f'{obj.__class__.__name__}({temp_v} ...)'
                rep_.append(f'{attr_} = {obj}')
                if len(rep_) >= attr_limit:
                    rep_.append('...')
                    break
        else:
            return repr(obj_)

    except Exception as err:
        logger.error(err)
        rep_ = []
    rep_ = ',\n    '.join(rep_)
    __repr_template = f"""
    {rep_}
    """
    return f"{obj_.__class__.__name__} ({__repr_template})"


def can_do(obj: Any) -> List[str]:
    """
    List methods and attributes of a Python object.

    It is essentially the builtin `dir` function without the private methods and attributes

    @param obj: Any
    @return: list of attr names sorted by name
    """
    return sorted([i for i in filter(lambda attr: not attr.startswith('_'), dir(obj))])


def sample(v, k=None, is_random=False) -> Union[list, dict, set, Any]:
    """
    Get sample of anything

    @param v: Any
    @param k: int
    @param is_random: default False
    @return: sample iterable
    """
    result = chunk(v, batch_size=k, is_random=is_random, max_batches=1)
    if len(result) == 1:
        result = result[0]
    if isinstance(v, set):
        result = set(result)
    return result


def type_table_of(o: Union[list, tuple, dict]):
    if isinstance(o, (list, tuple)):
        type_table = {i: type(i) for i in o}
    elif isinstance(o, dict):
        type_table = {}
        for k, v in o.items():
            if isinstance(o, dict):
                v = type_table_of(v)
            type_table[k] = (v, type(v))
    else:
        type_table = {o: type(o)}

    return type_table


def camel_to_snake(value: str):
    snaked_ = []
    for i, char in enumerate(value):
        if not i == 0 and char.isupper():
            char = f'_{char}'
        snaked_.append(char)
    return ''.join(snaked_).lower()


def get_implements(klass: type):
    classes = klass.__subclasses__()
    collected_classes = []
    for k in classes:
        k_classes = k.__subclasses__()
        if k_classes:
            collected_classes += get_implements(k)
        if not k.__name__.startswith('_'):
            collected_classes.append(k)
    return collected_classes


def get_instances_of(klass: type):
    return filter(lambda x: isinstance(x, klass), gc.get_objects())


def invert_dict(dict_: Union[dict, set]) -> dict:
    """
    Inverts the key by value
    e.g:
    >>> dict_ = {"a": "b", "c": "d"}
    >>> invert_dict(dict_)
    {"b" : "a", "d": "c"}
    :return: dict
    """

    if not isinstance(dict_, dict):
        raise TypeError("Send a dict object.")
    new_dict = {}
    for key, value in dict_.items():
        if isinstance(value, dict):
            new_dict.update({key: invert_dict(value)})
            continue
        if isinstance(value, (tuple, list, set)):
            new_dict.update({k: key for k in dict_[key]})
            continue
        new_dict[value] = key
    return new_dict


def import_string(dotted_path):
    """
    Import a dotted module path and return the attribute/class designated by the
    last name in the path. Raise ImportError if the import failed.
    """
    try:
        module_path, class_name = dotted_path.rsplit('.', 1)
    except ValueError as err:
        raise ImportError(f"{dotted_path} doesn't look like a module path") from err

    module = import_module(module_path)

    try:
        return getattr(module, class_name)
    except AttributeError as err:
        raise ImportError(f'Module {module_path} does not define a {class_name} attribute/class') from err


def get_attr_if_exists(obj: Any, attr: str) -> Union[object, None]:
    if hasattr(obj, attr):
        return getattr(obj, attr)
    return None


def time_format(seconds: float, format_='%H:%M:%S') -> Union[str, float]:
    """
    Default format is '%H:%M:%S'

    >>> time_format(3600)
    '01:00:00'

    """
    # this because NaN
    if seconds >= 0 or seconds < 0:
        time_ = time.strftime(format_, time.gmtime(abs(seconds)))
        if seconds < 0:
            return f"-{time_}"
        return time_
    return seconds  # NaN


def fill(value: Union[list, str, tuple], max_size, with_=' ') -> Any:
    """
    Calculates and adds value
    """
    fill_values = [with_] * (max_size - len(value))
    if isinstance(value, str):
        fill_values = ' '.join(fill_values)
        value = f"{value}{fill_values}"
    elif isinstance(value, list):
        value += fill_values
    elif isinstance(value, tuple):
        value += tuple(fill_values)
    return value


def list_methods(klass) -> List[str]:
    methods = []
    for i in dir(klass):
        if i.startswith('_') or not callable(getattr(klass, i)):
            continue
        methods.append(i)
    return methods


def string_to_literal(val: str):
    if isinstance(val, str):
        try:
            return ast.literal_eval(val)
        except:
            pass
    return val


def module_references(instance: types.ModuleType, **kwargs) -> dict:
    """
    dict of all functions and classes defined in the module.
    To also list the variables it is necessary to define explicitly with the special variable on your module
    _include
    **kwargs:
    _include -> to includes any definition and variables
    _exclude -> to exclude any definition
    :param instance:
    :return: List[str]
    """
    assert isinstance(instance, types.ModuleType), "You need to submit a module instance."
    logger.debug(f"Checking module {instance.__name__}")
    definitions = {}
    for i in dir(instance):
        if i.startswith('_'):
            continue
        exclude = get_attr_if_exists(instance, "_exclude") or kwargs.get("_exclude") or []
        include = get_attr_if_exists(instance, "_include") or kwargs.get("_include") or []

        obj = get_attr_if_exists(instance, i)

        if i in include:
            definitions[i] = obj

        if obj is not None and i not in exclude and callable(obj):
            if obj.__module__ == instance.__name__:
                definitions[i] = obj
    logger.debug(f"Collected: {definitions}")
    return definitions


def install_if_not(lib_name: str):
    from ..display import console
    try:
        importlib.import_module(lib_name)
        output = 'Alredy Installed'
    except ImportError:
        from ..system.commons import run_on_terminal

        command_ = f"{sys.executable} -m pip install {lib_name}"
        output = run_on_terminal(command_)
    console.log(output)


def set_log_level(level: Union[int, str]):
    """
    Default log level is INFO
    CRITICAL = 50
    FATAL = CRITICAL
    ERROR = 40
    WARNING = 30
    WARN = WARNING
    INFO = 20
    DEBUG = 10
    NOTSET = 0
    """
    log = logging.getLogger()
    log.setLevel(level)
    logger.info(f"Update log level to {level}")


def logger_level():
    import logging
    return logging.getLogger().level


def combine_with_all(a: list, b: list, n_a_combinations: int = 1, is_random: bool = False) -> List[Tuple[Any, ...]]:
    """
    >>> a = [1, 2, 3]
    >>> b = ['anything_a', 'anything_b']
    >>> combine_with_all(a, b)
    [(1, 'anything_a'), (1, 'anything_b'), (2, 'anything_a'), (2, 'anything_b'), (3, 'anything_a'), (3, 'anything_b')]
    >>> combine_with_all(a, b, n_a_combinations=2)
    [((1, 2), 'anything_a'), ((1, 2), 'anything_b'),
    ((1, 3), 'anything_a'), ((1, 3), 'anything_b'),
    ((2, 3), 'anything_a'), ((2, 3), 'anything_b')]
    """
    if not isinstance(n_a_combinations, int):
        raise TypeError(f"Please send {int}.")
    n_a_combinations = len(a) if n_a_combinations > len(a) else abs(n_a_combinations)

    combination = itertools.combinations(a, n_a_combinations) if n_a_combinations > 1 else a
    product_with_b = list(itertools.product(combination, b))
    if is_random:
        random.shuffle(product_with_b)
    return product_with_b


class CjTest(object):
    __template_unittest_function = """
    def test_{func_name}(self):
        pass
        """
    __template_unittest_class = """
class {class_name}Test(unittest.TestCase):
    {func_tests}
        """

    __template_unittest = """import unittest

{tests}

if __name__ == '__main__':
    unittest.main()

        """

    __prefix_attr_err = "Attr Check Error {attr_}."

    def __init__(self, instance_obj: object):
        self._prefix_attr = f"__{instance_obj.__class__.__name__}__"
        self._instance_obj = instance_obj
        self._set_attr_current_values()
        self._checks = []
        self._n_checks_passed = 0

    @property
    def checks(self):
        return self._checks

    @property
    def n_checks(self):
        return len(self._checks)

    @property
    def _instance_obj_attrs(self):
        return filter(lambda attr_: attr_.__contains__('__') is False, dir(self._instance_obj))

    def _get_attr_obj(self, attr_: str):
        if not hasattr(self._instance_obj, attr_):
            raise ValueError(f"Attr {attr_} not found.")
        value = getattr(self._instance_obj, attr_)
        return self._Attr(attr_, value)

    def _set_attr_current_values(self):
        for attr_ in self._instance_obj_attrs:
            attr_obj = self._get_attr_obj(attr_)
            attr_name = self.parse_attr(attr_)
            setattr(self, attr_name, attr_obj)

    def parse_attr(self, attr_: str):
        attr_ = self._valid_attr(attr_)
        return f'{self._prefix_attr}{attr_}'

    def __getattr__(self, item):
        return self.__getattribute__(self.parse_attr(item))

    class _Attr(object):
        def __init__(self, name: str, value: Any):
            self.name = name
            self.is_callable = callable(value)
            self.is_private = self.name.startswith('_')
            self.is_bool = value is True or value is False
            self.is_class = isinstance(value, ClassType)
            self.is_function = isinstance(value, FunctionType)
            self.class_of_attr = value.__class__
            self._operator_repr = None
            self.tests_case = []

        def __repr__(self):
            return f"{self.name}"

        def __str__(self):
            return f"{self.name}"

        def __len__(self):
            return len(self.tests_case)

        def __eq__(self, other):
            """ ==value """
            if isinstance(other, self.__class__):
                return NotImplemented
            self.tests_case = (other, self.class_of_attr.__eq__, '==')
            return self

        def __ge__(self, other):
            """>=value """
            if isinstance(other, self.__class__):
                return NotImplemented
            self.tests_case = (other, self.class_of_attr.__ge__, '>=')
            return self

        def __gt__(self, other):
            """>value """
            if isinstance(other, self.__class__):
                return NotImplemented
            self.tests_case = (other, self.class_of_attr.__gt__, '>')
            return self

        def __le__(self, other):
            """ <=value. """
            if isinstance(other, self.__class__):
                return NotImplemented
            self.tests_case = (other, self.class_of_attr.__le__, '<=')
            return self

        def __lt__(self, other):
            """ <value. """
            if isinstance(other, self.__class__):
                return NotImplemented
            self.tests_case = (other, self.class_of_attr.__lt__, '<')
            return self

        def __ne__(self, other):
            """ !=value. """
            if isinstance(other, self.__class__):
                return NotImplemented
            self.tests_case = (other, self.class_of_attr.__ne__, '!=')
            return self

        def copy(self):
            return copy(self)

        def run(self, current_value):
            expected, operator, _ = self.tests_case
            if not operator(current_value, expected):
                return [f"{repr(current_value)} not {_} {repr(expected)}"]

            return []

    def _valid_attr(self, attr_name: str):
        assert hasattr(self._instance_obj,
                       attr_name), f"{self.__prefix_attr_err.format(attr_=repr(attr_name))} isn't defined."
        return attr_name

    def add_check(self, *check_: _Attr):
        for i in check_:
            if i not in self._checks:
                self._checks.append(i)

    def remove_check(self, index: int):
        self._checks.pop(index)

    def check_attr(self, attr_name: Union[str, _Attr]):
        if isinstance(attr_name, str):
            stored_test = self.__getattribute__(self.parse_attr(attr_name))
        else:
            stored_test = attr_name
        current_value = getattr(self._instance_obj, stored_test.name)
        if not stored_test.is_callable:
            tests_ = stored_test.run(current_value)
            passed = not any(tests_)
            self._n_checks_passed += len(stored_test) - len(tests_)
            msg_err = f"{self.__prefix_attr_err.format(attr_=repr(stored_test.name))} {' '.join(tests_)}"
            assert passed, msg_err

    def check_all(self):
        for attr_ in self._checks:
            self.check_attr(attr_)

    @classmethod
    def _get_class_test(cls, ref):
        func_tests = ''.join(cls.__template_unittest_function.format(func_name=i) for i in list_methods(ref))
        return cls.__template_unittest_class.format(class_name=ref.__name__, func_tests=func_tests)

    @classmethod
    def _get_func_test(cls, ref):
        return cls.__template_unittest_function.format(func_name=ref.__name__)

    @classmethod
    def _get_test(cls, ref):
        if isinstance(ref, (FunctionType, types.MethodType)):
            return cls._get_func_test(ref)
        if isinstance(ref, type):
            return cls._get_class_test(ref)
        raise TypeError("send a function or class reference")

    @classmethod
    def build_test(cls, reference):
        module_func_test = []
        tests = []
        if isinstance(reference, types.ModuleType):
            for _, ref in module_references(reference).items():
                if isinstance(ref, type):
                    tests.append(cls._get_test(ref))
                    continue
                module_func_test.append(cls._get_test(ref))
        else:
            if isinstance(reference, type):
                tests.append(cls._get_test(reference))
            else:
                module_func_test.append(cls._get_test(reference))
        if module_func_test:
            module_func_test = ''.join(module_func_test)
            tests = [cls.__template_unittest_class.format(class_name='Module', func_tests=module_func_test)] + tests
        return cls.__template_unittest.format(tests='\n'.join(tests))


def _add_license(base_dir, ext='.py'):
    from cereja.file import FileIO
    from cereja.config import BASE_DIR
    licence_file = FileIO.load(BASE_DIR)
    for file in FileIO.load_files(base_dir, ext=ext, recursive=True):
        if 'Copyright (c) 2019 The Cereja Project' in file.string:
            continue
        file.insert('"""\n' + licence_file.string + '\n"""')
        file.save(exist_ok=True)


def _rescale_down(input_list, size):
    assert len(input_list) >= size, f'{len(input_list), size}'

    skip = len(input_list) // size
    for n, i in enumerate(range(0, len(input_list), skip), start=1):
        if n > size:
            break
        yield input_list[i]


def _rescale_up(values, k, fill_with=None):
    size = len(values)
    assert size <= k, f'Error while resizing: {size} < {k}'
    clones = (math.ceil(abs(size - k) / size))
    for i in values:
        vals = (i,) + ((fill_with,) if fill_with is not None else (i,)) * clones
        for val in vals:

            k -= 1
            if k < 0:
                break
            yield val


def rescale_values(values: List[Any], granularity: int, **kwargs) -> List[Any]:
    """
    Resizes a list of values
    eg.
        >>> import cereja as cj
        >>> cj.rescale_values(values=list(range(100)), granularity=12)
        [0, 8, 16, 24, 32, 40, 48, 56, 64, 72, 80, 88]
        >>> cj.rescale_values(values=list(range(5)), granularity=10)
        [0, 0, 1, 1, 2, 2, 3, 3, 4, 4]


    @param values: Sequence of anything
    @param granularity: is a integer
    @param kwargs:
        fill_with: Any value
    @return: rescaled list of values.
    """

    if len(values) >= granularity:
        result = list(_rescale_down(values, granularity))
    else:
        result = list(_rescale_up(values, granularity, **kwargs))

    assert len(result) == granularity, f"Error while resizing the list size {len(result)} != {granularity}"
    return result


class Source:

    def __init__(self, reference: Any):
        self._source_code = inspect.getsource(reference)

    @property
    def source_code(self):
        return self._source_code

    def save(self, path_, **kwargs):
        from cereja import FileIO, Path
        assert Path(path_).suffix == '.py', "Only python source code."
        FileIO.create(path_, self._source_code).save(**kwargs)


def is_iterable(obj: Any) -> bool:
    """
    Return whether an object is iterable or not.

    :param obj: Any object for check
    """
    return isinstance(obj, Iterable)


def is_sequence(obj: Any) -> bool:
    """
    Return whether an object a Sequence or not, exclude strings and empty obj.

    :param obj: Any object for check
    """
    return not isinstance(obj, (str, dict, bytes)) and is_iterable(obj)


def is_numeric_sequence(obj: Sequence[Number]) -> bool:
    try:
        from cereja.array import flatten
        sum(flatten(obj))
    except (TypeError, ValueError):
        return False
    return True



import tempfile
import unittest
from abc import abstractmethod

from cereja.system import Path
from cereja.file import FileIO
from cereja.display import console


class FileIOTest(unittest.TestCase):
    data = None

    @property
    def name(self):
        return self.__class__.__name__

    @abstractmethod
    def create(self):
        file = FileIO.create('test', data=self.data)  # generic
        return file

    def setUp(self) -> None:
        self.file = self.create()

    def test_sanity(self):
        console.log(f'Testing {self.name}')
        with tempfile.TemporaryDirectory() as tempdir:
            self.file.set_path(Path(tempdir).join(self.file.name))
            data_before_save = self.file.data
            self.file.save(exist_ok=True)
            file = FileIO.load(self.file.path)
            self.assertTrue(file.path.exists, msg="File don't exist.")
            self.assertEqual(file.data, data_before_save, msg='Data corrupted')
            file.delete()
            self.assertFalse(file.path.exists)


class FileIOTxtTest(FileIOTest):
    data = ['first line', 'second line', 'third line']

    def create(self):
        file = FileIO.txt('test.txt', data=self.data, creation_mode=True)
        return file

    def test_history(self):
        for line in self.file:
            pass

        self.assertEqual(self.file[0], 'first line')
        self.assertEqual(self.file[:3], ['first line', 'second line', 'third line'])
        # Insert Data
        self.file.add('other line', line=0)
        self.file.add('other line2', line=0)
        self.assertEqual(self.file.data, ['other line2', 'other line', 'first line', 'second line', 'third line'])
        # it is allowed to use index assignment
        self.file[0] = 'other line'

        self.file.add('test')  # append
        self.assertEqual(self.file.data,
                         ['other line', 'other line', 'first line', 'second line', 'third line', 'test'])

        # Data Recovery
        self.file.undo()  # You selected amendment 4
        self.file.undo()  # You selected amendment 3
        self.assertEqual(self.file.data, ['other line2', 'other line', 'first line', 'second line', 'third line'])
        self.file.redo()  # You selected amendment 4
        self.assertEqual(self.file.data,
                         ['other line', 'other line', 'first line', 'second line', 'third line'])


class FileIOJsonTest(FileIOTest):
    data = {'key': 'value', 'key2': 'value2', 'key3': 'value3'}

    def create(self):
        return FileIO.json('test.json', data=self.data, creation_mode=True)

    def test_commons_operations(self):
        self.assertEqual(self.file.data, {'key': 'value', 'key2': 'value2', 'key3': 'value3'})

        # Iterable
        for key, value in self.file.items():
            pass

        self.assertEqual(self.file['key'], 'value')

        # Insert Data
        self.file['key4'] = 'value4'
        self.assertEqual(self.file.data, {'key': 'value', 'key2': 'value2', 'key3': 'value3', 'key4': 'value4'})

    def test_sample_items(self):
        self.assertEqual(self.file.sample_items(), {'key': 'value'})
        self.assertEqual(self.file.sample_items(k=3), {'key': 'value', 'key2': 'value2', 'key3': 'value3'})
        self.assertRaises(AssertionError, self.file.sample_items, -1)
        self.assertRaises(AssertionError, self.file.sample_items, 0.5)


class FileCsvTest(FileIOTest):
    data = [[1, 2, 3]]
    cols = ['col1', 'col2', 'col3']

    def create(self):
        return FileIO.csv('test.csv', data=self.data, creation_mode=True, cols=self.cols)

    def test_commons_operations(self):
        self.assertEqual(self.file.data, [[1, 2, 3]])
        self.file.add([1, 2], fill_with=0)
        self.assertEqual(self.file.data, [[1, 2, 3], [1, 2, 0]])

        # convert to dict
        self.assertEqual(self.file.to_dict(), {'col1': (1, 1), 'col2': (2, 2), 'col3': (3, 0)})

        # or get generation row by row with col
        self.assertEqual(list(self.file.rows), [{'col1': 1, 'col2': 2, 'col3': 3}, {'col1': 1, 'col2': 2, 'col3': 0}])

        # Iterable
        for row in self.file:
            pass

        # indexing col values
        self.assertEqual(self.file['col1'], (1, 1))
        # or use index and get a row
        self.assertEqual(self.file[0], [1, 2, 3])
        self.assertEqual(self.file.flatten(), [1, 2, 3, 1, 2, 0])


class FilePyObjTest(FileIOTest):
    data = [[1, 2, 3]]

    def create(self):
        return FileIO.create('test.pkl', data=self.data)

    def test_commons_operations(self):
        self.assertEqual(self.file.data, [[1, 2, 3]])


if __name__ == '__main__':
    unittest.main()



import tempfile
import unittest

from cereja.mltools import preprocess

from cereja.system import Path
from cereja.mltools.preprocess import separate
from cereja.mltools.data import Tokenizer
from cereja.mltools.data import Freq


class BaseTest(unittest.TestCase):

    def loop_on_examples_pairs(self, func, examples):
        msg_ = f"Error on {func.__name__}"
        for example, expected_response, args, kwargs in examples:
            if args and kwargs:
                self.assertEqual(func(example, *args, **kwargs), expected_response, msg=msg_)
            elif args:
                self.assertEqual(func(example, *args), expected_response, msg=msg_)
            elif kwargs:
                self.assertEqual(func(example, **kwargs), expected_response, msg=msg_)
            else:
                self.assertEqual(func(example), expected_response, msg=msg_)


class TokenizerTest(unittest.TestCase):
    data = None

    def test_sanity(self):
        with tempfile.TemporaryDirectory() as tempdir:
            to_encode = 'oi tudo tranquilo'
            tokenizer = Tokenizer(['oi amigo tudo bem'], preprocess_function=separate, use_unk=False)

            self.assertEqual(tokenizer.last_index, 13)
            encoded = tokenizer.encode([to_encode])
            valid_encode_tokens = [[tokenizer._item_to_index['oi'], tokenizer._item_to_index['tudo'],
                                    tokenizer._item_to_index['tranquilo']]]
            self.assertEqual(tokenizer.last_index, 14)
            self.assertEqual(encoded, valid_encode_tokens)
            decoded = ' '.join(tokenizer.decode(encoded[0]))
            self.assertEqual(decoded, to_encode)
            temp_file = Path(tempdir).join('tokenizer.json')
            tokenizer.to_json(temp_file)

            # Valid.
            tokenizer = Tokenizer.load_from_json(temp_file)
            encoded = tokenizer.encode([to_encode])
            self.assertEqual(tokenizer.last_index, 14)
            self.assertEqual(encoded, valid_encode_tokens)
            decoded = ' '.join(tokenizer.decode(encoded[0]))
            self.assertEqual(decoded, to_encode)


class PreprocessModuleTest(BaseTest):

    def test_sanity(self):
        self.assertEqual(preprocess.separate('how are you?', sep='?'), 'how are you ?')
        self.assertEqual(preprocess.separate('how are you,man?', sep=('?', ','), between_char=True),
                         'how are you , man ?')
        self.assertEqual(preprocess.separate('how are! you?'), 'how are ! you ?')

    def test_separate(self):
        examples = (
            ('how are you?', 'how are you ?', (), {'sep': '?'}),
            ('how are you,man?', 'how are you , man ?', (), {'sep': ('?', ','), 'between_char': True}),
            ('how are! you?', 'how are ! you ?', (), {}),
        )
        self.loop_on_examples_pairs(preprocess.separate, examples)

    def test_accent_remove(self):
        examples = (('olá tudo bem amigo? Você é um cara legal!', 'ola tudo bem amigo? Voce e um cara legal!', (), {}),
                    ('isso é muito difícil', 'isso e muito dificil', (), {}),
                    )
        self.loop_on_examples_pairs(preprocess.accent_remove, examples)

    def test_replace_english_contractions(self):
        examples = (
            ("i'll", "i shall", (), {}),
        )
        self.loop_on_examples_pairs(preprocess.replace_english_contractions, examples)

    def test_remove_extra_chars(self):
        examples = (('i   like it!!!', 'i like it!', (), {}),)
        self.loop_on_examples_pairs(preprocess.remove_extra_chars, examples)

    def test_remove_non_language_elements(self):
        examples = (('\U0001F352 i like it!', 'i like it!', (), {}),)
        self.loop_on_examples_pairs(preprocess.remove_non_language_elements, examples)

    def test_remove_punctuation(self):
        examples = (('\U0001F352 i like it!', '\U0001F352 i like it', (), {}),
                    ('olá tudo bem amigo?! Você é um cara legal!', 'olá tudo bem amigo Você é um cara legal', (), {}),
                    ('e aí? quero R$ 40,00', 'e aí quero R 4000', (), {}),
                    )
        self.loop_on_examples_pairs(preprocess.remove_punctuation, examples)

        examples2 = (('\U0001F352 i like it!', '\U0001F352 i like it', (), {}),
                     ('olá tudo bem amigo?! Você é um cara legal!', 'olá tudo bem amigo Você é um cara legal', (), {}),
                     ('e aí? quero R$ 40,00', 'e aí quero R$ 40,00', (), {'punctuation': '?!'}),
                     )
        self.loop_on_examples_pairs(preprocess.remove_punctuation, examples2)

    def test_remove_stop_words(self):
        examples = (('my name is Joab', 'my name Joab', (), {}),
                    ('you are cool!', 'you cool!', (), {}),
                    )
        self.loop_on_examples_pairs(preprocess.remove_stop_words, examples)


class FreqTest(unittest.TestCase):
    def setUp(self) -> None:
        self.freq = Freq([1, 2, 3, 4, 4])

    def test_probability(self):
        self.assertEqual(self.freq.probability, {1: 0.2, 2: 0.2, 3: 0.2, 4: 0.4})

    def test_item_prob(self):
        self.assertEqual(self.freq.item_prob(4), 0.4)

    def test_sanity(self):
        self.assertEqual(preprocess.separate('how are you?', sep='?'), 'how are you ?')
        self.assertEqual(preprocess.separate('how are you,man?', sep=('?', ','), between_char=True),
                         'how are you , man ?')
        self.assertEqual(preprocess.separate('how are! you?'), 'how are ! you ?')

        freq = Freq([1, 1, 2, 2, 3, 3, 4, 5, 6, 7, 8, 'hi', 'o', 'a'])
        self.assertEqual(freq.sample(max_freq=1), {4: 1, 5: 1, 6: 1, 7: 1, 8: 1, 'hi': 1, 'o': 1, 'a': 1})
        self.assertEqual(freq.sample(freq=2), {1: 2, 3: 2, 2: 2})

        self.assertRaises(AssertionError, freq.sample, freq=1, max_freq=2)
        self.assertRaises(AssertionError, freq.sample, freq=1, min_freq=2)

        freq = Freq([1, 2, 3, 3, 4, 5, 6, 7, 6, 7, 12, 31, 123, 5, 3])
        self.assertEqual(freq.least_freq(), {123: 1, 31: 1, 12: 1, 4: 1, 2: 1, 1: 1, 7: 2, 6: 2, 5: 2, 3: 3})


if __name__ == '__main__':
    unittest.main()



import threading
import time

import random


def test(value):
    time.sleep(random.randint(1, 10))
    return value


class Process:
    def __init__(self, function):
        self._function = function
        self._result = []
        self._ths = {}
        self._stop_on = None
        self._current = 0

    def _wrap(self, *args, id_=None, **kwargs):
        self._result.append((id_, self._function(*args, **kwargs)))
        self._current += 1

    def run_on(self, values, **kwargs):
        self._stop_on = len(values)
        for n, value in enumerate(values, 1):
            th = threading.Thread(target=self._wrap, args=(value,), kwargs={'id_': n, **kwargs})
            th.start()
            self._ths[n] = th
        while self._current < self._stop_on:
            if len(self._result) >= 1:
                id_, val = self._result.pop(0)
                self._ths[id_].join()
                yield val


process = Process(test)
for i in process.run_on(range(100)):
    print(i)



"""

Copyright (c) 2019 The Cereja Project

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from .config import conf
from cereja.utils import *
from . import utils
from . import display
from cereja.display import *
from . import file
from cereja.file import *
from . import array
from cereja.array import *
from . import system
from cereja.system import *
from cereja.system.unicode import *
from . import utils
from .utils import decorators
from cereja.concurrently import *
from . import mltools
from cereja.mltools import *
from cereja.utils.version import get_version_pep440_compliant
from cereja.system.unicode import *
from . import date
from cereja.date import *
from . import hashtools
from . import mathtools
from cereja.mathtools import *
from . import experimental
from ._requests import request

VERSION = "1.4.4.final.0"

__version__ = get_version_pep440_compliant(VERSION)

NON_BMP_SUPPORTED = None
if NON_BMP_SUPPORTED is None:
    # This is important, as there may be an exception if the terminal does not support unicode bmp
    try:
        unicode_ = f"\033[31m\U0001F352\033[30m"
        print(f"{unicode_} Using Cereja v.{get_version_pep440_compliant()}\r")
        NON_BMP_SUPPORTED = True
    except (UnicodeEncodeError, UnicodeDecodeError, UnicodeError, UnicodeTranslateError):
        NON_BMP_SUPPORTED = False
# cj_modules_dotted_path = utils.import_string('cereja.conf.cj_modules_dotted_path')
#
# for dot_module in cj_modules_dotted_path:
#     globals().update(utils.module_references(import_module(dot_module)))



"""

Copyright (c) 2019 The Cereja Project

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import logging
import os
import random
import shutil
import time
import warnings
from datetime import datetime
from typing import List, Union

from cereja.array import group_items_in_batches
from pathlib import Path as Path_


logger = logging.getLogger(__name__)

__all__ = ['Path', 'change_date_from_path', 'clean_dir', 'file_name', 'get_base_dir', 'group_path_from_dir', 'listdir',
           'mkdir', 'normalize_path']


def mkdir(path_dir: str):
    """
    Creates directory regardless of whether full cereja exists.
    If a nonexistent cereja is entered the function will create the full cereja until it creates the requested directory.
    e.g:
    original structure
    content/

    after create_dir(/content/cereja/to/my/dir)
    content/
        --cereja/
            --to/
                --my/
                    --dir/

    It's a recursive function

    :param path_dir: directory cereja
    :return: None
    """

    dir_name_path = os.path.dirname(path_dir)
    if not os.path.exists(dir_name_path):
        mkdir(dir_name_path)
    if not os.path.exists(path_dir):
        os.mkdir(path_dir)


def group_path_from_dir(dir_path: str, num_items_on_tuple: int, ext_file: str, to_random: bool = False,
                        key_sort_function=None):
    """
    returns data tuples based on the number of items entered for each tuple, follows the default order
    if no sort function is sent

    :param dir_path: full path to dir
    :param num_items_on_tuple: how much items have your tuple?
    :param ext_file: is file extension to scan
    :param to_random: randomize result
    :param key_sort_function: function order items
    :return:
    """

    if '.' not in ext_file:
        ext_file = '.' + ext_file

    key_sort_function = {"key": key_sort_function} if key_sort_function else {}
    paths = sorted([os.path.join(dir_path, i) for i in os.listdir(dir_path) if ext_file == os.path.splitext(i)[1]],
                   **key_sort_function)

    batches = group_items_in_batches(items=paths, items_per_batch=num_items_on_tuple)

    if to_random:
        random.shuffle(batches)
    return batches


def file_name(file_path: str, with_ext: bool = False) -> str:
    """
    Pass a path from a file and the file name will be returned

    >>> file_name('/content/myfile.ext')
    'myfile'
    >>> file_name(file_path='/content/myfile.ext', with_ext=True)
    'myfile.ext'

    :param file_path: string with file path
    :param with_ext: define if you want the file name with the extension
    :return: the file name
    """

    if not isinstance(file_path, str):
        raise TypeError('Path must be a string')

    base_name = os.path.basename(file_path)
    if with_ext:
        return base_name
    f_name, _ = os.path.splitext(base_name)
    return f_name


def change_date_from_path(path_: str, format_: str = '%d-%m-%Y %H:%M:%S'):
    m_time = os.path.getctime(path_)
    return time.strftime(format_, time.localtime(m_time))


def get_base_dir(__file__: str) -> str:
    """
    :param __file__: send your __file__ string.
    :return: absolute path.
    """
    return os.path.normpath(os.path.dirname(os.path.abspath(__file__)))


def _norm_path(path_: str):
    return os.path.normpath(path_) if os.path.isabs(path_) else os.path.abspath(path_)


class Path(os.PathLike):
    _date_format = "%Y-%m-%d %H:%M:%S"
    __sep = os.sep

    def __init__(self, initial: Union[str, os.PathLike] = '.', *pathsegments: str):
        self.__path = Path_(_norm_path(initial), *pathsegments)
        self.__parent = self.__path.parent.as_posix()
        self._parent_name = self.__path.parent.name
        self._verify()

    def __fspath__(self):
        return str(self.__path)

    def __repr__(self):
        return self.path

    def __str__(self):
        return self.path

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            other = self.__class__(other)
        return self.path == other.path

    def __len__(self):
        return len(self.parts)

    def __contains__(self, item):
        return item in self.path

    def __add__(self, other):
        return self.join(*other)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __getitem__(self, item):
        if item < 0:
            raise ValueError(f'index < 0 is not possible.')
        if isinstance(item, int):
            item = slice(item + 1)
        return self.__class__('/'.join(self.parts[item]))

    def __value_err(self, details=''):
        if details:
            details = f'details {details}'
        raise ValueError(f"Path <{self.path}> isn't valid. {details}")

    def _verify(self):
        part = self.__path
        for i in range(len(self.__path.parts)):
            part_split = part.name.split('.')
            if part_split[-1] == '.':
                logger.info(f'It is not common to use dot <{part.name}> in the end of name.')
                break
            if (part.suffix or part_split[-1] == '.') and i > 0:
                logger.info(f'It is not common to use dot in the middle or end of directory name <{part.name}>')
                break
            if len(part_split) > 2:
                logger.info(f"<{part.name}> has more dot than usual.")
            part = part.parent

    @property
    def name(self):
        return self.__path.name

    @property
    def exists(self):
        return self.__path.exists()

    @property
    def path(self):
        return self.__path.as_posix()

    @property
    def stem(self):
        return self.__path.stem

    @property
    def suffix(self):
        return self.__path.suffix

    @property
    def root(self):
        return self.__path.anchor

    @property
    def parent(self):
        """
        :return: Parent PathInfo object
        """
        return self.__class__(self.__parent)

    @property
    def parent_name(self):
        return self._parent_name

    @property
    def uri(self):
        try:
            return self.__path.as_uri()
        except:
            return ''

    @property
    def is_dir(self):
        return self.__path.is_dir()

    @property
    def is_file(self):
        return self.__path.is_file()

    @property
    def is_link(self):
        return self.__path.is_symlink()

    @property
    def updated_at(self):
        return datetime.fromtimestamp(os.stat(str(self.path)).st_mtime).strftime(
                self._date_format) if self.exists else None

    @property
    def created_at(self):
        return datetime.fromtimestamp(os.stat(str(self.path)).st_ctime).strftime(
                self._date_format) if self.exists else None

    @property
    def last_access(self):
        return datetime.fromtimestamp(os.stat(str(self.path)).st_atime).strftime(
                self._date_format) if self.exists else None

    @property
    def is_hidden(self):
        """
        In Unix-like operating systems, any file or folder that starts with a dot character
        (for example, /home/user/.config), commonly called a dot file or dotfile,is to be
        treated as hidden – that is, the ls command does not display them unless the -a
        flag ( ls -a ) is used.

        Return True if the name starts with dot or path contains one of the special names reserved
        by the system, if any.

        :return: bool
        """
        return self.name.startswith('.') or self.__path.is_reserved()

    @property
    def parts(self):
        return self.__path.parts

    @property
    def sep(self):
        return self.__sep

    @classmethod
    def get_current_dir(cls) -> 'Path':
        """
        Get current working directory
        @return:
        """
        return cls(os.getcwd())

    @classmethod
    def change_current_dir(cls, to_):
        os.chdir(to_)

    def join(self, *args):
        assert self.__path.suffix == '', f"join operation is only dir. full path received {self.path}"
        return self.__class__(self.__path.joinpath(*args).as_posix())

    def _shutil(self, command: str, **kwargs):
        if not self.exists:
            raise Exception(f'Not Found <{self.uri}>')
        if command == 'rm':
            rm_tree = kwargs.get('rm_tree')
            if self.is_file:
                os.remove(self.path)
                logger.info(f"<{self}> has been removed")
            else:
                if rm_tree is True:
                    shutil.rmtree(str(self))
                    logger.info(f"<{self}> has been removed")
                else:
                    try:
                        os.rmdir(str(self))
                    except OSError as err:
                        raise Exception(f'{err}.\n Use rm_tree=True to DELETE {self.uri}.')
        elif command == 'mv':
            to = kwargs.get('to')
            shutil.move(str(self), str(to))
            return to
        elif command == 'cp':
            to = kwargs.get('to')
            if self.is_dir:
                shutil.copytree(str(self), str(to))
            elif self.is_file:
                shutil.copy(str(self), str(to))
            return to

    def rm(self, rm_tree=False):
        """
        [!] Use caution when using this command. Only use if you are sure of what you are doing.

        If it is a directory you can DELETE the entire tree, for that flag `rm_tree=True`
        """
        return self._shutil('rm', rm_tree=rm_tree)

    def mv(self, to):
        to = self.__class__(to)
        return self._shutil('mv', to=to)

    def cp(self, to):
        to = self.__class__(to)
        return self._shutil('cp', to=to)

    def rsplit(self, sep=None, max_split=-1):
        return self.path.rsplit(sep, max_split)

    def split(self, sep=None, max_split=-1):
        return self.path.rsplit(sep, max_split)

    @classmethod
    def list_dir(cls, path_: Union[str, 'Path'], only_name=False) -> List['Path']:
        """
        Extension of the listdir function of module os.

        The difference is that by default it returns the absolute path, but you can still only request the relative path by
        setting only_relative_path to True.

        """
        if not isinstance(path_, Path):
            path_ = cls(path_)

        assert path_.is_dir, f"{path_} isn't dir."
        return [p.stem if only_name else p for p in map(path_.join, os.listdir(path_))]

    @classmethod
    def list_files(cls, dir_path: Union[str, 'Path'], ext: str = None, contains_in_name: List = (),
                   not_contains_in_name=(),
                   recursive=False) -> List['Path']:

        ext = ext or ''
        files = []
        for p in cls.list_dir(dir_path):
            if p.is_dir and recursive:
                files.extend(
                        cls.list_files(p, ext=ext, contains_in_name=contains_in_name,
                                       not_contains_in_name=not_contains_in_name,
                                       recursive=recursive))
                continue
            if p.is_dir:
                continue
            if not (p.suffix == f'.{ext.strip(".")}' or ext == ''):
                continue
            if not_contains_in_name:
                if any(map(p.stem.__contains__, not_contains_in_name)):
                    continue
            if contains_in_name:
                if not any(map(p.stem.__contains__, contains_in_name)):
                    continue
            files.append(p)
        return files


def normalize_path(path_: str) -> Path:
    return Path(path_)


def clean_dir(path_: str):
    """
    Delete all files on dir
    """
    content = Path.list_dir(path_)
    for p in content:
        p.rm(rm_tree=True)
    logger.info(f"{len(content)} files were removed")


def listdir(path_: Union[str, Path], only_relative_path: bool = False) -> List[str]:
    """
    Extension of the listdir function of module os.

    The difference is that by default it returns the absolute path, but you can still only request the relative path by
    setting only_relative_path to True.

    """
    warnings.warn(f"This function will be deprecated in future versions. "
                  f"Path.list_dir", DeprecationWarning, 2)
    if not isinstance(path_, Path):
        path_ = Path(path_)

    assert path_.is_dir, f"{path_} isn't dir."
    return [path_.join(p) if not only_relative_path else p for p in os.listdir(str(path_))]


if __name__ == '__main__':
    pass
